##############################################################################
# Example Python script that uses the ML adapter API in Mascot Server.       #
##############################################################################
# COPYRIGHT NOTICE                                                           #
# Copyright (C) 2025 Matrix Science Limited  All Rights Reserved.            #
#                                                                            #
# This script is free software. You can redistribute and/or                  #
# modify it under the terms of the GNU General Public License                #
# as published by the Free Software Foundation; either version 2             #
# of the License or, (at your option), any later version.                    #
#                                                                            #
# These modules are distributed in the hope that they will be useful,        #
# but WITHOUT ANY WARRANTY; without even the implied warranty of             #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the               #
# GNU General Public License for more details.                               #
#                                                                            #
# Original author: Ville Koskinen                                            #
##############################################################################

import sys
import logging
import argparse
import csv

import msparser

# Format of log messages.
LOGGING_TMPL = u'%(asctime)-15s %(levelname)s %(message)s'

# Exit codes for this script.
EXIT_SUCCESS = 0
EXIT_FAILURE = 1

logging.basicConfig(format=u'%(asctime)-15s %(levelname)s %(message)s', level=logging.INFO)

MascotDat = msparser.ms_datfile('../config/mascot.dat')

if not MascotDat.isValid():
    logging.error("This program must be run from the mascot/bin directory, because it needs to read options from config/mascot.dat. %s", MascotDat.getLastErrorString())
    sys.exit(1)

def _create_argparse() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        formatter_class=argparse.RawTextHelpFormatter,
        description="""Read a query list and compute a feature for each PSM.

example_ML_Adapter.py is an example Python script that uses the ML adapter
API in Mascot Server. For demonstration purposes, the script computes the
peptide length as a Percolator feature.

The script receives a query list file as input. Rows in this file with
label=1 ask for predicted data for target PSMs; with label=-1, decoy
PSMs. The script looks up the corresponding target and decoy PSMs in the
results file using Mascot Parser.
"""
    )

    parser.add_argument(
        "--resfile",
        required=True,
        help="""Path to the Mascot results file (F001234.dat, F001234.msr).
This is the file from which PSMs are loaded based on the query and
rank in the query list file.
""",
    )

    parser.add_argument(
        "--query_list",
        required=True,
        help="""Path to the query list input file.
This is a tab-separated (TSV) file containing the columns:

query    rank    label

Query and rank refer to Mascot query and rank of the peptide-spectrum
match. Label is 1 (target) or -1 (decoy).
""",
    );

    parser.add_argument(
        "--output",
        required=True,
        help="""Predicted values are written in this filepath.
This is a tab-separated (TSV) file containing columns:

  query    rank    label    [feature 1]    [feature 2] ...

The names and number of feature columns depend on the adapter.
"""
    )

    parser.add_argument(
        'other_arguments',
        nargs="*",
        help=''
    )

    return parser


class QueryRankLabel():
    """Simple data structure for storing query,rank,label and PSMid."""

    def __init__(self, query : str, rank : str, label : str):
        self.query = query
        self.rank = rank
        self.label = label

    def __eq__(self, other):
        return self.query == other.query and self.rank == other.rank and self.label == other.label

    def __hash__(self):
        return hash((self.query, self.rank, self.label))

def read_query_list(fh) -> list[QueryRankLabel]:
    """Read a query list input file and return a list of ids.

    The file is a tab-separated (TSV) file containing the columns:

    query    rank    label
    
    """

    reader = csv.DictReader(fh, delimiter="\t")
    columns = reader.fieldnames

    PSMids = []

    for desired in ['query', 'rank', 'label']:
        if desired not in columns:
            logging.error("Expected a '{}' column but got '{}'".format(desired, columns))
            return PSMids

    for cells in reader:
        PSMids.append(QueryRankLabel(int(cells['query']), int(cells['rank']), int(cells['label'])))

    return PSMids
    

def open_resfile(mascot_options : msparser.ms_mascotoptions, filepath : str) -> msparser.ms_mascotresfilebase:
    """Open the Mascot results file at filepath. Please check
    isValid() and getLastErrorString() on the returned object for any
    warnings or errors.

    If the file is not a supported file type, the function returns an
    object that inherits from ms_mascotresfilebase but isn't a valid
    results file ('ms_mascotresfile_nil').

    mascot_options should come from ms_datfile.getMascotOptions().

    """

    keepaliveInterval = 1
    keepalive = """ul=Creating unassigned list (%p% complete)
qu=Calculating quantitation component intensities (%p% complete)
ci=Creating cache file (%p% complete)
rf=Reading results file (%p% complete)
ap=Assigning peptides to proteins (%p% complete) hit=%h, time=%d
gp=Found protein group: %a, hit=%h, %p% complete, %d seconds
cc=Caching results (%p% complete)
fd=Calculating false discovery rate (%p% complete)
Opening results file %p%
    """

    flags = msparser.ms_mascotresfilebase.RESFILE_USE_CACHE
    cache_directory = mascot_options.getCacheDirectory()

    return msparser.ms_mascotresfilebase.createResfile(filepath, keepaliveInterval, keepalive, flags, cache_directory)

def define_pepsum_params(mascot_options : msparser.ms_mascotoptions, resfile : msparser.ms_mascotresfilebase, is_decoy : bool) -> msparser.ms_mascotresults_params:
    """Set flags and options for ms_peptidesummary. These are based on:

    the defaults from ms_mascotresfilebase.get_ms_mascotresults_params();
    whether the search is error tolerant;
    whether it's a spectral library search;
    whether it's an intact crosslink search;
    and whether Mascot options specify that caching should be used.

    mascot_options should come from ms_datfile.getMascotOptions().

    resfile should come from open_resfile()

    """

    params = msparser.ms_mascotresults_params()

    # Use the default options configured in mascot.dat.
    resfile.get_ms_mascotresults_params(mascot_options, params)

    # Enable or disable caching.
    params.setFlag2(msparser.ms_peptidesummary.MSPEPSUM_USE_CACHE, True)

    # We don't need protein inference because we are only accessing
    # PSM data. This special flag speeds up opening the file, but make
    # sure you heed Parser documentation:
    #
    #   "Used for when a ms_peptidesummary object it required, but no
    #   protein grouping is needed.
    #   ms_peptidesummary::getNumberOfHits() will return zero. The
    #   only functions that are guaranteed to work are
    #   ms_peptidesummary::getPeptide(),
    #   ms_peptidesummary::getAllProteinsWithThisPepMatch(),
    #   ms_peptidesummary::getQmatch(). Will not work with error
    #   tolerant searches."
    params.setFlag2(msparser.ms_peptidesummary.MSPEPSUM_NO_PROTEIN_GROUPING, True)

    # Target or decoy matches?
    params.setFlag(msparser.ms_mascotresults.MSRES_DECOY, is_decoy)

    # Don't want to enable Percolator mode. Parser should not try to
    # read the pop files in an ML adapter.
    params.setFlag2(msparser.ms_peptidesummary.MSPEPSUM_PERCOLATOR, False)

    return params

def compute_features(query_list : list[QueryRankLabel], pepsum_target : msparser.ms_peptidesummary, pepsum_decoy : msparser.ms_peptidesummary) -> tuple[list[str], dict[QueryRankLabel, list]]:
    """Compute the features for each PSM in the query list. Return the
    list of feature names and a dictionary of features for each PSM.

    The function loads the corresponding ms_peptide object from
    pepsum_target or pepsum_decoy, then computes the peptide length.
    If the PSM doesn't exist, it assigns peptide length zero.

    """

    # Feature name is an arbitrary string that can be anything that is
    # relevant to the ML adapter. The Mascot core feature set also has
    # peptideLength, but you can use the same name here. When the
    # computed/predicted features are merged with core features, they
    # are prefixed with the adapter name, so there can never be name
    # ambiguity.
    feature_columns = ['peptideLength']
    features = {}

    for QRL in query_list:
        peptide = None

        if QRL.label == 1:
            peptide = pepsum_target.getPeptide(QRL.query, QRL.rank)
        elif QRL.label == -1:
            peptide = pepsum_decoy.getPeptide(QRL.query, QRL.rank)

        if peptide:
            logging.debug("Fetched ms_peptide object for query={} rank={} label={}".format(QRL.query, QRL.rank, QRL.label))

            # The computation in this case is trivial. However, you
            # could use arbitrary Python libraries to compute or
            # predict features or interact with external AI/ML
            # services over HTTP. All the PSM data is available
            # through ms_peptide and ms_peptidesummary, and all the
            # peak lists are available through ms_inputquery.
            peplen = len(peptide.getPeptideStr())

            logging.debug("Computed peptideLength={}".format(peplen))

            features[QRL] = [peplen]

        else:
            logging.warning("Failed to find ms_peptide object for query={} rank={} label={}; imputing 0 as peptideLength".format(QRL.query, QRL.rank, QRL.label))

            features[QRL] = [0]

    return feature_columns, features

def write_features(filepath : str, query_list : list[QueryRankLabel], feature_columns : list[str], features : dict[QueryRankLabel, list]):
    """Write the features in the output file. It is a tab-separated
    (TSV) file containing columns:

    query    rank    label    [feature 1]    [feature 2] ...

    """

    n = 0

    with open(filepath, 'w', newline='', encoding='utf-8') as f:
        header = ['query', 'rank', 'label']
        header.extend(feature_columns)

        csvwriter = csv.writer(f, delimiter='\t')
        csvwriter.writerow(header)

        for QRL in query_list:
            row = [QRL.query, QRL.rank, QRL.label]

            if QRL in features:
                row.extend(features[QRL])

            csvwriter.writerow(row)
            n = n + 1

    logging.info("Wrote data for {} PSMids in output file.".format(n))

def main():
    parser = _create_argparse()
    args = parser.parse_args()

    sys.stdout.reconfigure(encoding='utf-8')
    sys.stderr.reconfigure(encoding='utf-8')

    with open(args.query_list, 'r', encoding='utf-8') as f:
        query_list = read_query_list(f)

    if not query_list:
        logging.error("Cannot proceed with empty query list.")
        sys.exit(EXIT_FAILURE)

    logging.info("Read {} items from query list file.".format(len(query_list)))

    # If the input is an MSR file, caching is unnecessary and the
    # createResfile call returns immediately. If the input is a .dat
    # file, caching is recommended. In this case, it's also
    # recommended to set output keepalive, so that periodic progress
    # is printed in stdout.
    #
    # See the Parser documentation for
    # ms_mascotresfilebase::createResfile().

    resfile = open_resfile(MascotDat.getMascotOptions(), args.resfile)

    if not resfile.isValid():
        logging.error(resfile.getLastErrorString())
        sys.exit(EXIT_FAILURE)

    # We also need the target peptide summary report and decoy peptide
    # summary report.
    pepsum_params_target = define_pepsum_params(MascotDat.getMascotOptions(), resfile, is_decoy=False)
    pepsum_target = msparser.ms_peptidesummary(resfile, pepsum_params_target)

    # Any errors are logged in the resfile object -- see Parser
    # documentation for details.
    if not resfile.isValid():
        logging.error(resfile.getLastErrorString())
        sys.exit(EXIT_FAILURE)

    pepsum_params_decoy = define_pepsum_params(MascotDat.getMascotOptions(), resfile, is_decoy=True)
    pepsum_decoy = msparser.ms_peptidesummary(resfile, pepsum_params_decoy)

    if not resfile.isValid():
        logging.error(resfile.getLastErrorString())
        sys.exit(EXIT_FAILURE)

    feature_columns, computed_features = compute_features(query_list, pepsum_target, pepsum_decoy)

    write_features(args.output, query_list, feature_columns, computed_features)

    sys.exit(EXIT_SUCCESS)

if __name__ == "__main__":
    main()
